%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Diagnostic Medical Image Processing (DMIP)
% WS 2014/15
% Author: Marco Boegel & Yan Xia
% Exercise 1: SVD & FT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% start Matlab: ./local/matlab/bin/matlab&
% or: addpackage matlab && matlab&

clear variables; % delete all used variables
close all; % close all applications
clc % clear the Command Window




%%
A = [ 11 10 14; 12 11 -13; 14 13 -66 ];
sprintf('det(A): %.2f', det(A))

[ U S V ] = svd(A)
U*S*V'


% Compute the inverse without the typical MATLAB command inv
Si = zeros(size(S));
numRows = size(S,1);
numColumns = size(S,2);

% How many diagonal elements do we have?
n = min([numRows, numColumns]);

% Invert the singular values
for d=1:n
    if ( S(d,d) > eps )
        Si(d,d) = 1.0 / S(d,d);
    else
        Si(d,d) = 0.0;
    end
end

Ainv = V * Si * U'



% Check using the inv command from MATLAB
disp('inv(A)');
inv(A)

disp('cond(A)');
cond(A)
disp('rank(A)');
rank(A)
disp('null(A)');
null(A)

% rank deficiency, if e=10^3
S(3,3) = 0;
A2 = U*S*V'
rank(A2)
null(A2)
U
V











%%
% Exercise problem:
% Show that a variation in b by 0.1% implies a change in x by 240%.

% Assume we have to solve the linear problem
% A x = b, where x is unknown
disp('A x = b');
b = [1.001 0.999 1.001]'
% Solution of the system
x = Ainv * b;
x

% If we round vector b we get a modification of 0.1% in each element of b
disp('br = round(b): A xr = br');
br = round(b)
xr = Ainv * br

% (We want only the difference caused by the 0.1% change)
% bn = br-b: Therefore, we get: A*xr=br = A*(x+xn)=b+bn
disp('bn = br-b: A xn = bn');
bn = br-b
xn = Ainv * bn

% Since bn indicates the modification of 0.1% in b we can express the 
% change in x by
disp('Modification of x');
xn./x
% Alternatively
% (xr-x)./x

% and see a modification in the components of x of [-2.4641 -2.1862
% -1.0000] each, which is around 240% for the first two components














%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Optimization problem I
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Let us consider the following problem that appears in many image
% processing and computer vision problems:
% We computed a matrix A out of sensor data like an image. By theory
% the matrix A must have the singular values s1, s2, . . . , sp, where
% p = min(m, n). Of course, in practice A does not fulfill this constraint.
% Problem: What is the matrix A0 that is closest to A (according to the
% Frobenius norm) and has the required singular values?

% see: 'help norm' - 'fro' The Frobenius-norm of matrix A, 
% sqrt(sum(diag(A'*A))).

% Let us assume that by theoretical arguments the matrix A is required
% to have a rank deficiency of one, and the two non-zero singular values
% are identical. The matrix A0 that is closest to A according to the
% Frobenius norm and fulfills the above requirements is:

% Build mean of first two singular values to make them identical
% Set third singular value to zero
nd = [ (S(1,1)+S(2,2))/2; (S(1,1)+S(2,2))/2; 0 ];
A0 = U * diag(nd) * V';
norm(A,'fro')
norm(A0,'fro')









%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Optimization problem IV
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The Moore-Penrose pseudo-inverse is required to find the
% solution to the following optimization problem:
% Compute the regression line through the following 2-D points:
% We have seven points (x_i,y_i) and want to fit a line model
% y = mx + t (linear approximation) through the points
% y = AA*b

AA = [ 3 1; 2 1; 1 1; 0 1; -1 1; -1 1; -2 1 ];
b = [ 2 1 2 0 1 -1 -1 ]';

% Moore-Penrose pseudo-inverse using: pinv
% Properties of the Moore-Penrose pseudo-inverse B = pinv(A):
% A*B*A = A
% B*A*B = B
% A*B is Hermitian
% B*A is Hermitian

l = pinv(AA)*b; % Least-square solution [l(1)=m, l(2)=t]
% Note: l = inv(AA)*b does not work, because AA is not square. We do have
% more samples (7) than unknowns (2) -> overdetermined system 
% -> least-square solution.

% Plot points
figure(4);
plot(AA(:,1), b(:), '.');
hold on;
x = -3:0.1:4;
% y = mx+t
y = l(1) * x + l(2);
plot(x,y,'.r');
title('2-D points (x-coord.) with an estimated line model');
hold off;











%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Optimization problem II
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Estimate the matrix A in R^{2,2} such that for the following vectors
% First step: Enter data as given on lecture slides
b1 = [ 1  1 ];
b2 = [-1  2 ];
b3 = [ 1 -3 ];
b4 = [-1 -4 ];

b = [1 -1  1 -1;
     1  2 -3 -4];
no = size(b,2)
MM = zeros(no,4);

for i=1:no
    MM(i,1) = b(1,i)^2;
    MM(i,2) = b(1,i)*b(2,i);
    MM(i,3) = MM(i,2);
    MM(i,4) = b(2,i)^2;
end


b2 = [ 1  1;
      -1  2;
       1 -3;
      -1 -4
      ];
  
no2 = size(b2,1);
MMM = zeros(no2,4);

for s=1:no2
    MMM(s,1) = b2(s,1)^2;
    MMM(s,2) = b2(s,1)*b2(s,2);
    MMM(s,3) = MMM(s,2);
    MMM(s,4) = b2(s,2)^2;
end

% the optimization problem gets solved:
% sum_{i=1}^{4} b'_{i} A b_{i} and ||A||_{F} = 1
% The objective function is linear in the components of A, thus the whole
% sum can be rewritten in matrix notation:
% Ma = 0
% where the measurement matrix M is built from single elements of the
% sum.
% Given the four points we get four equations and therefore four rows in M:

M = [ 1 1 1 1; 1 -2 -2 4; 1 -3 -3 9; 1 4 4 16 ];

% TASK: estimate the matrix A
% SOLUTION: Compute vector that spans the nullspace of A; we can find it in
% the last column vector of V.

[ U S V ] = svd(M);
null(M)
a = V(:,end)

% Reshape array, returns a m-by-n matrix B whose elements are taken
% column-wise from a
A = reshape(a,2,2)';
A
norm(A,'fro') % check ||A||_{F} = 1
a'*M*a % check solution (-> min)







%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Optimization problem III
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The SVD can be used to compute the image matrix of rank 1 that best
% approximates an image. Figure 1 shows an example of an image I
% and its rank-1-approximation I0 = u_{1} * s_{1} * v'_{1}.

xray1 = imread('yu_fill.jpg');
xray = double(xray1);
figure(1);
subplot(1,4,1);
imagesc(xray);
axis image;
colormap(gray)
title('Original image');
axis image
% Label of the x-axis
xlabel('x');
% Label of the y-axis
ylabel('y');


% Create the SVD of the image
[U S V] = svd(xray);
% Approximate image of rank 1
I1 = U(:,1) * S(1,1) * V(:,1).';


subplot(1,4,2);
colormap(gray)
imagesc(I1)
title('Rank 1 Approximation');
axis image
% Label of the x-axis
xlabel('x');
% Label of the y-axis
ylabel('y');


subplot(1,4,3);
colormap(gray)
imagesc(xray ./ I1);
axis image;
title('Corrected: xray ./ I1');
axis image
% Label of the x-axis
xlabel('x');
% Label of the y-axis
ylabel('y');


% TASK: Compute rank k approximation
% SOLUTION:
k = 100;
Id = zeros(size(xray));
% create a vector v for plotting
v = zeros(1,k);

% uses getframe and movie2avi to create a movie
figure(2);
for d = 1:k
    Id = Id + U(:,d) * S(d,d) * V(:,d).';
    v(1,d) = S(d,d);
    imagesc(Id);
    axis image;
    colormap(gray);
    title(['Rank ', int2str(d), ' Approximation']);
    xlabel('x');
    ylabel('y');
    France(d) = getframe(gcf);
end
movie2avi(France,'rank.avi','fps',5);

figure(3);
x = 1:k;
plot(x, v(1,x),'-r');
xlabel('Number of the singular value');
ylabel('Singular value');
title('Distribution of the singular values');

figure(1);
subplot(1,4,4);
colormap(gray)
imagesc(Id);
axis image;
title(['Rank ', int2str(k), ' Approximation']);
axis image
% Label of the x-axis
xlabel('x');
% Label of the y-axis
ylabel('y');

% Nice to know!
print -deps 'svd.eps'
print -djpeg 'svd.jpeg'
saveas(1, 'svd', 'tif')

% choose rank k approximation
rankImage = zeros(size(xray));
variable = input('Enter the chosen rank: ')
rankImage = rankImage + U(:,1:variable) * S(1:variable,1:variable)...
    * V(:,1:variable).';

figure(6);
colormap(gray)
imagesc(rankImage);
axis image;
title(['Rank ', int2str(variable), ' Approximation']);
axis image
% Label of the x-axis
xlabel('x');
% Label of the y-axis
ylabel('y');







%%
%%%%%%%%%%%%%%%%%%%%%
% Fourier Transform %
%%%%%%%%%%%%%%%%%%%%%

% Create a head phantom image, 'Modified Shepp-Logan',
% number of rows and columns should be 8 / 128
p = phantom(128);
figure(5);
subplot(2,2,1);
imagesc(p);
axis image;
colormap gray
title 'Phantom image'
axis image
% Apply the 2-D discrete Fourier transform
% Puts center of the beam in the corners
I = fft2(p);
% The DC component of the signal is in the upper left corner
% of the 2-D FFT
sum(sum(p))
subplot(2,2,2);
imagesc(abs(I));
axis image;
%title '|fft2(p)|'
axis image
size(I)
subplot(2,2,3);
% Shift zero-frequency component to center of spectrum
% Shift DC component of Fast Fourier transform to center of spectrum
% Put center of the beam into the center
imagesc(abs(fftshift(I)));
axis image;
%title '|fftshift(fft2(p))|'
axis image
Z=fftshift(I)
% m rows, n columns
[m n] = size(p);
% The DC component is near the center of the matrix
Z(m/2, n/2)
subplot(2,2,4);
imagesc(log(1+abs(fftshift(I))));
axis image;
%title 'log(1+|fftshift(fft2(p))|)'
axis image